from enum import Enum, auto
from typing import Optional, Iterable, Dict

from dataclasses import dataclass


class Engine(Enum):
    ADA = auto()
    BABBAGE = auto()
    CURIE = auto()
    DAVINCI = auto()
    INSTRUCT_CURIE_BETA = auto()
    INSTRUCT_DAVINCI_BETA = auto()

    def __str__(self):
        return self.name.lower().replace('_', '-')


@dataclass
class CompletionConfig:
    max_tokens: int = 10
    temperature: float = 0.7
    top_p: float = 0.0
    n: int = 1
    #stream: bool = False
    logprobs: Optional[int] = None
    echo: bool = False
    stop: Optional[Iterable[str]] = None
    presence_penalty: int = 0
    frequency_penalty: int = 0
    best_of: int = 1
    #logit_bias: Optional[Dict[str, int]] = None



class DefaultConfig:
    engine = Engine.ADA
    max_chars = 7500
    completion_config = CompletionConfig()




from dataclasses import dataclass, asdict
from typing import Iterable

import openai
from openai.openai_object import OpenAIObject

from goldpot.Config import DefaultConfig, CompletionConfig, Engine
from goldpot.Model import CompletionInput, CompletionOutput, SearchInput, SearchOutput


@dataclass
class Goldpot:
    api_key: str
    engine: Engine = DefaultConfig.engine
    max_chars: int = DefaultConfig.max_chars
    completion_config: CompletionConfig = DefaultConfig.completion_config

    def __post_init__(self):
        openai.api_key = self.api_key

    def complete_(self, prompt: str) -> OpenAIObject:
        return openai.Completion.create(engine=str(self.engine), prompt=prompt, **asdict(self.completion_config))

    def complete(self, input: CompletionInput) -> CompletionOutput:
        prompts = input.get_prompts()
        return CompletionOutput([self.complete_(prompt) for prompt in prompts], prompts)

    def search_(self, query: str, docs: Iterable[str]) -> OpenAIObject:
        return openai.Engine(str(self.engine)).search(documents=docs, query=query)

    def search(self, input: SearchInput) -> SearchOutput:
        return SearchOutput(
            [self.search_(q := query_docs[0], ds := query_docs[1]) for query_docs in input.get_docs()], input.docs)



import re
from dataclasses import dataclass, field
from typing import Optional, Iterable, Any, Dict, Tuple

from openai.openai_object import OpenAIObject

from goldpot import DefaultConfig
from goldpot.Utils import Splitter


@dataclass
class CompletionInput:
    start: str = ''
    end: str = ''
    max_chars: int = DefaultConfig.max_chars
    sep: str = '\n'
    sample_sep: str = '----'
    samples: Iterable[str] = field(default_factory=lambda: [])

    def get_prompts(self):
        samples = ''.join([self.sep + sample + self.sep + self.sample_sep + self.sep for sample in self.samples])
        begin = (self.start + samples)[:self.max_chars]
        left_chars = self.max_chars - len(begin)
        splits = Splitter.len_chunk_split(re.findall('[\w\W]*?[.!?]|[\w\W]+', self.end), left_chars)
        return [begin + ''.join(s) for split in splits for s in split]


@dataclass
class SearchInput:
    query: str
    docs: [str]
    max_chars: int = DefaultConfig.max_chars

    def get_docs(self):
        left_chars = self.max_chars - len(self.query)
        docs_split = Splitter.len_chunk_split(self.docs, left_chars)
        return [(self.query, d) for docs in docs_split for d in docs]


@dataclass
class CompletionResult:
    prompt: str
    finish_reason: str
    text: str
    logprobs: Any


@dataclass
class CompletionOutput:
    output: Iterable[OpenAIObject]
    prompts: Iterable[str]
    completion: str = ''
    completions: Iterable[str] = field(default_factory=lambda: [])
    results: Iterable[CompletionResult] = field(default_factory=lambda: [])

    def __post_init__(self):
        self.results = [CompletionResult(oo[1], c.finish_reason, c.text, c.logprobs) for oo in
                        zip(self.output, self.prompts) for c in oo[0].choices]
        self.completions = [r.text for r in self.results]
        self.completion = '\n'.join(self.completions)


@dataclass
class SearchOutput:
    output: Iterable[OpenAIObject]
    docs: Iterable
    scores: Dict[str, float] = field(default_factory=lambda: {})
    scores_sorted: Iterable[Tuple[str, float]] = field(default_factory=lambda: [])

    def __post_init__(self):
        clip = lambda f: max(min(300.0, f), 0.0) / 300.0
        r = [(o.score, clip(o.score)) for oo in self.output for o in oo.data]
        self.scores = dict(zip(self.docs, r))
        self.scores_sorted = sorted(self.scores.items(), key=lambda i: i[1][1], reverse=True)


import itertools
import sys
from itertools import takewhile, accumulate
from typing import Iterable


class Splitter:
    # TODO: more general (2d) knapsack opt approach
    # TODO: tail rec for current implementation

    @staticmethod
    def len_chunk_split(items: Iterable[str], max_len: int = sys.maxsize, max_chunks: int = sys.maxsize) -> Iterable[str]:

        def chunk(iterable, size):
            it = iter(iterable)
            while item := list(itertools.islice(it, size)):
                yield item

        def rec(items, res: Iterable[str] = [], pending=True):
            if len(items) == 0 or not pending:
                return res
            else:
                split_n = len(list(takewhile(lambda cum_len: not cum_len > max_len, accumulate([len(s) for s in items]))))
                return rec(items[split_n:], res + [items[:split_n]], split_n > 0)

        return [list(chunk(split, max_chunks)) for split in rec(items) if len(split) > 0]